#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 * This small utility generates UV coordinates for characters on
 * a texture.
 *
 * Characters are expected to be monospace, aligned in rows from
 * top to bottom.
 *
 * The coordinate system puts (0;0) at the bottom-left of the
 * texture.
 *
 * (ptx1; pty1)---(ptx2; pty1)
 * |               |
 * |               |
 * |               | The coordinates of a caracter.
 * |   character   | The ouput will be: ptx1 pty1 ptx2 pty1 ptx1 pty2 ptx2 pty2
 * |               |
 * |               |
 * |               |
 * |               |
 * (ptx1; pty2)---(ptx2; pty2)
 *
 */

typedef struct Vec2{
	float x, y;
}Vec2;

//this was written with dimensions in pixels in mind.
Vec2 chr;
Vec2 img;
int chrn = -1;
int verbose = 0;

void printinfo(void);
int getres(Vec2 *, char *);


int
main(int argc, char **argv)
{
	int len;
	int ok = 0;

	for(int i = 1; i < argc; i++){
		if(argv[i][0] == '-'){
			int step = 0;

			len = strlen(argv[i]);
			for(int e = 1; e < len; e++){
				switch(argv[i][e]){
				case 'v':
					verbose = 1;
					break;
				case 'h':
					printinfo();
					return 0;

				case 'n': /*number of characters*/
					step++;
					if(i+step >= argc){
						//uh oh, no more args to get
						printinfo();
						return -1;
					}
					if(sscanf(argv[i+step], "%d", &chrn) != 1){
						fprintf(stderr, "err: %s not understood\n", argv[i+step]);
						return 1;
					}
					chrn = abs(chrn);
					break;

				case 's': /*character size*/
					step++;
					if(i+step >= argc){
						printinfo();
						return 1;
					}

					if(getres(&chr, argv[i+step])<0){
						fprintf(stderr, "err: %s not understood\n", argv[i+step]);
						return 1;
					}
					ok++;
					break;

				case 'S': /*image/texture size*/
					step++;
					if(i+step >= argc){
						printinfo();
						return 1;
					}

					if(getres(&img, argv[i+step])<0){
						fprintf(stderr, "err: %s not understood\n", argv[i+step]);
						return 1;
					}
					ok++;
					break;

				default:
					printinfo();
					return 1;
				}
			}
			i+=step; //to not get args that were already parsed in the switch
		}
	}

	if(ok<2){
		fprintf(stderr, "err: mandatory arguments were not given\n");
		printinfo();
		return 1;
	}



	chr.x /= img.x;
	chr.y /= img.y;

	if (verbose)
		printf("char height: %f, char width:  %f\n\
img height: %f, img width: %f\n\
nbr of chars to print: %d\n", chr.x, chr.y, img.x, img.y, chrn);

	float ptx1 = 0;
	float ptx2 = chr.x;
	float pty1 = 1;
	float pty2 = 1 - chr.y;

	int i = 0;
	int line = 0;

	while (pty2 >= 0) { //unnecessary check just in case

		if (verbose)
			printf("line %d\n", ++line);


		//print the uv coordinates of all the chars on the same line
		//uv X coordinates go from 0 to 1 (left to right)
		while ( ptx2 <= 1) {
			if (verbose)
				printf("CHAR %d : ", i+1);
			printf("%f %f %f %f %f %f %f %f\n",
				ptx1, pty1,
				ptx2, pty1,
				ptx1, pty2,
				ptx2, pty2);

			//increment the x values by the width of the char
			ptx1 += chr.x;
			ptx2 += chr.x;

			i++;
			//check if we've alredy done all the chars
			if (chrn != -1 && i >= chrn) {
				if (verbose)
					printf("Printed %d characters over %d line(s)\n", i, line);
				return 0;
			}
		}


		//if we get here, it means en entire line has been filled.
		//decrement the y values by the height of the char.
		//uv Y coordinates go from 1 to 0 (top to bottom)
		pty1 -= chr.y;
		pty2 -= chr.y;

		//reset the position of the x values (we will begin a new line)
		ptx1 = 0;
		ptx2 = chr.x;


		//if the lower y value goes outside of the range [1,0], it means we aren't in
		//the texture anymore
		if (pty2 < 0) {
			if(chrn != -1){
				fprintf(stderr, "err: the image's dimensions aren't big enough\n");
				return 1;
			}
			if(verbose)
				printf("Printed %d characters over %d line(s)\n", i, line);
			return 0;
		}
	}
	return 0;
}

int
getres(Vec2 *res, char *str)
{
	if(sscanf(str, "%fx%f", &res->x, &res->y) != 2)
		return -1;

	//I could simply get the absolute value, but I don't want to link with -lm
	if(res->x <= 0 || res->y <= 0)
		return -1;
	return 0;
}

void
printinfo(void)
{
	fprintf(stderr, "Usage: uvcalc [-hvn] -s XxY -S XxY\n\n\
-h\n\
	Display this help and quit.\n\
-v\n\
	Enable verbose mode.\n\
-s\n\
	Define characters' width and height. Example: -s 11x21\n\
-S\n\
	Define the image's width and height. Example: -S 512x512\n\
-n\n\
	Set the number of characters. Without this option, characters will \
be generated until the image is full.\n\n\
Example: uvcalc -vsSn 11x21 512x512 224\n");
}
